"""Load SDFITS files produced by the Green Bank Telescope"""
import sys
import copy
from astropy.wcs import WCS
from astropy.units import cds
from astropy.io import fits
from astropy.modeling import models, fitting
import astropy.units as u
from astropy.table import Table
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from ..spectra.spectrum import Spectrum
from ..spectra.obsblock import Obsblock
from ..spectra import dcmeantsys
from .sdfitsload import SDFITSLoad
from ..util import uniq

class GBTFITSLoad(SDFITSLoad):
    """GBT-specific container for bintables from selected HDU(s)"""
    def __init__(self, filename, source=None,hdu=None):
        SDFITSLoad.__init__(self,filename,source,hdu)#,fix=False)
        print("==GBTLoad %s" % filename)

        self.ushow(0,'OBJECT')
        self.ushow(0,'SCAN')
        self.ushow(0,'SAMPLER')
        #ushow('PLNUM')
        #ushow('IFNUM')
        self.ushow(0,'SIG')
        self.ushow(0,'CAL')
        self.ushow(0,'PROCSEQN')
        self.ushow(0,'PROCSIZE')
        self.ushow(0,'OBSMODE')  
        self.ushow(0,'SIDEBAND')




from specutils import Spectrum1D, SpectrumList,SpectralRegion
from astropy.modeling.polynomial import Polynomial1D,Chebyshev1D
from astropy.modeling.fitting import LevMarLSQFitter,LinearLSQFitter
from astropy.wcs import WCS
from . import baseline

class Spectrum(Spectrum1D):
    """basic spectrum built on `~specutils.Spectrum1D`
     with added attributes like baseline model

    """

    def __init__(self, *args,**kwargs):
        Spectrum1D.__init__(self,*args,**kwargs)
        self._baseline_model = None

    @property
    def baseline_model(self):
        return self._baseline_model

    def baseline(self,order,exclude=None,**kwargs):
        """compute and optionally remove a baseline"""
        kwargs_opts = {
            'remove':False,
            'model':'polynomial',
            'fitter':  LinearLSQFitter(calc_uncertainties=True),
        }
        kwargs_opts.update(kwargs)

        self._baseline_model = baseline(self,order,exclude,**kwargs)
        if kwargs_opts['remove']:
            s=self.subtract(self._baseline_model(self.spectral_axis))
            self._data = s._data

    def _undo_baseline(self):
        s=self.add(self._baseline_model(self.spectral_axis))
        self._data = s._data
        self._baseline_model = None
    
    def bshow(self):
        print(f"baseline model {self._baseline_model}")


    def stats(self):
        """ Return some stats. Note this works with slicing!"""
        mean = self.mean()
        rms = self.data.std()
        dmin = self.min()
        dmax = self.max()
        return (mean,rms,dmin,dmax)

    def smooth(self):
        #todo use specutils.manipulation.smoothing
        #option to smooth baseline too?
        pass




import sys
import copy
import numpy as np
from astropy.wcs import WCS
from astropy.units import cds
from astropy.modeling.models import Gaussian1D
from astropy.modeling.polynomial import Polynomial1D
from astropy.io import fits
from astropy.modeling import models, fitting
import astropy.units as u
from astropy.table import Table
from astropy.modeling.fitting import LevMarLSQFitter,LinearLSQFitter
from specutils import Spectrum1D, SpectrumList,SpectralRegion
from specutils.fitting import fit_continuum
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from .spectrum import Spectrum

class Obsblock():
    '''Class that holds a series of spectra on which bulk operations can be performed'''
    def __init__(self,speclist,index):
        self._speclist = speclist
        self._index = index # pandas dataframe

    def __getitem__(self,i):
        return self._speclist[i]

    def __len__(self):
        return len(self._speclist)

    def __op__(self,opname):
        pass

    def baseline(self,order,exclude=None,**kwargs):
        """compute and optionally remove a baseline"""
        pass




import numpy as np
from specutils import Spectrum1D, SpectrumList, SpectralRegion
from specutils.fitting import fit_continuum
import astropy.units as u
from astropy.modeling.models import Gaussian1D
from astropy.modeling.polynomial import Polynomial1D
from memory_profiler import profile
import matplotlib.pyplot as plt
import sys
import resource
import gc
import objgraph

#@profile
def testit(nchan,num):
    # Make no difference if use spectrumlist or []
    if False:
        sl = SpectrumList()
    else:
        sl = []
    #default size 
    raw = np.random.rand(nchan)
    rawsize=sys.getsizeof(raw)
    blocksize=rawsize*num
    x = np.linspace(0,10,nchan)
    g1 = Gaussian1D(amplitude=10,mean=5,stddev=0.3)
    print(f"Size of single {nchan} channel spectrum {rawsize/(1024*1024):.2} Mbytes")
    print(f"X {num} = {blocksize} bytes = {blocksize/(1048576*1024):.2} GB")
    for j in range(num):
        if j%500 == 0:
            print(f"doing row {j}")
        sp = (20.+g1(x)+np.random.normal(size=nchan))*u.K
        p=Spectrum1D(flux=sp, spectral_axis=x*u.um,wcs=None,meta={},velocity_convention="doppler_radio")
        sl.append(p)
    return sl

def baseline(speclist,order,exclude=None,plot=False):
    fc = fit_continuum(p,Polynomial1D(degree=order),exclude_regions=[exclude])
    if plot:
        fig,ax = plt.subplots()
        ax.plot(x,p.flux)
        ax.plot(x,fc(x))
        plt.show()

if __name__ == "__main__":
    sl=testit(nchan=2048,num=2000)
    print(f'Memory usage: %s (Gb) {resource.getrusage(resource.RUSAGE_SELF).ru_maxrss/(1024*1024)}')
    gc.collect()
    print(f'Post-GC Memory usage: %s (Gb) {resource.getrusage(resource.RUSAGE_SELF).ru_maxrss/(1024*1024)}')
    objgraph.show_most_common_types()
    p = sl[0]
    x = sl[0].spectral_axis
    nchan = len(x)
    cdelt = x[1]-x[0]
    center = x[nchan // 2]
    width = 0.25*nchan*cdelt
    exclude = SpectralRegion.from_center(center,width)
    i=0
    for sp in sl:
        if (i%500) == 0:
            print("baseline ",i)
        baseline(sp,1,exclude)
        i=1+i



#!/usr/bin/env python

"""Tests for `astrodysh` package."""

import subprocess

from astrodysh.astrodysh import astrodysh


def test_cli():
    """Ensure that the CLI has been installed"""
    subprocess.check_call(["astrodysh", "foo", "bar"])


def test_hello_world():
    assert astrodysh("foo", "bar") is True



import sys
import copy
import numpy as np
from astropy.wcs import WCS
#from astropy.modeling.models import Gaussian1D
from astropy.io import fits
#from astropy.modeling import models, fitting
import astropy.units as u
from astropy.table import Table
from astropy.modeling.polynomial import Polynomial1D,Chebyshev1D
from astropy.modeling.fitting import LevMarLSQFitter,LinearLSQFitter
from specutils import Spectrum1D, SpectrumList,SpectralRegion
from specutils.fitting import fit_continuum
import matplotlib.pyplot as plt
#from scipy.optimize import leastsq

def baseline_all(speclist,order,exclude=None,**kwargs):
    kwargs_opts = {
        'remove': False,
        'show': False,
        'model':'polynomial',
        'fitter':  LinearLSQFitter(calc_uncertainties=True),
    }
    kwargs_opts.update(kwargs)
    i=0
    bad = 0
    for p in speclist:
        p.baseline(order,exclude,**kwargs)

def baseline(spectrum,order,exclude=None,**kwargs):
    kwargs_opts = {
        #'remove': False,
        #'show': False,
        'model':'polynomial',
        'fitter':  LinearLSQFitter(calc_uncertainties=True),
    }
    kwargs_opts.update(kwargs)

    _valid_models = ["polynomial", "chebyshev"]
    # @todo replace with minimum_string_match
    if kwargs_opts["model"] not in _valid_models:
        raise ValueError(f'Unrecognized input model {kwargs["model"]}. Must be one of {_valid_models}')
    if kwargs_opts['model'] == "polynomial":
        model = Polynomial1D(degree=order)
    elif kwargs_opts['model'] == "chebyshev":
        model = Chebyshev1D(degree=order)
    else:
        # should never get here
        raise ValueError(f'Unrecognized input model {kwargs["model"]}. Must be one of {_valid_models}')
    fitter = kwargs_opts['fitter']
    #print(f"MODEL {model} FITTER {fitter}")
    p = spectrum
    if np.isnan(p.data).all():
        print('returning none')
        return None # or raise exception
    if exclude is not None:
        return fit_continuum(spectrum=p,
                model=model,
                fitter=fitter,
                exclude_regions=[exclude])
    else:
        return fit_continuum(spectrum=p,
                model=model,
                fitter=fitter)

def baseline_old(speclist,order,exclude=None,**kwargs):
    kwargs_opts = {
        'remove': True,
        'show': False,
        'model':'polynomial',
        'fitter':  LinearLSQFitter(calc_uncertainties=True),
    }
    kwargs_opts.update(kwargs)

    _valid_models = ["polynomial", "chebyshev"]
    # @todo replace with minimum_string_match
    if kwargs["model"] not in _valid_models:
        raise ValueError(f'Unrecognized input model {kwargs["model"]}. Must be one of {_valid_models}')
    print(f"BL {order} for {len(speclist)} spectra")
    i=0
    bad = 0
    if kwargs_opts['model'] == "polynomial":
        model = Polynomial1D(degree=order)
    elif kwargs_opts['model'] == "chebyshev":
        model = Chebyshev1D(degree=order)
    else:
        # should never get here
        raise ValueError(f'Unrecognized input model {kwargs["model"]}. Must be one of {_valid_models}')
    fitter = kwargs_opts['fitter']
    print(f"MODEL {model} FITTER {fitter}")
    try:
        if exclude is not None:
            for p in speclist:
                if np.isnan(p.data).all():
                    bad+=1
                    continue
                fc = fit_continuum(spectrum=p,
                        model=model,
                        fitter=fitter,
                        exclude_regions=[exclude])
                i=i+1
        else:
            for p in speclist:
                if np.isnan(p.data).all():
                    bad+=1
                    continue
                fc = fit_continuum(spectrum=p,
                        model=model,
                        fitter=fitter)
                i=i+1
    except Exception as e:
        print(f"At spectrum {i}, Exception was {e}")
        print(p)
        print("DATA MEAN: ",np.nanmean(p.data))
        return p
    if plot:
        fig,ax = plt.subplots()
        ax.plot(x,p.flux)
        ax.plot(x,fc(x))
        plt.show()
    print(f"NUMBER OF BAD SPECTRA: {bad}")
    return None

def dcmeantsys(calon, caloff, tcal, mode=0, fedge=10, nedge=None):
    """
    following the GBTIDL routine with same name, get the tsys from 
    the neighboring calon and caloff we define an extra way to set 
    the edge size, nedge, if you prefer to use number of edge channels
    instead of the inverse fraction
    
    calon/caloff is meant to reflect the state of the noise diode
    
    mode=0     do the mean before the division
    mode=1     do the mean after the division
    """
    nchan = len(calon)
    if nedge == None:
        nedge = nchan // fedge     # 10 %
    if mode == 0:
        meanoff = np.mean(caloff[nedge:-nedge])
        meandiff = np.mean(calon[nedge:-nedge] - caloff[nedge:-nedge])
        meanTsys = ( meanoff / meandiff * tcal + tcal/2.0 )
    else:
        meanTsys = np.mean( caloff[nedge:-nedge] / (calon[nedge:-nedge] - caloff[nedge:-nedge]) )
        meanTsys = meanTsys * tcal + tcal/2.0
    return meanTsys



"""
Core utility classes and functions
"""
import numpy as np

def get_size(obj, seen=None):
    """Recursively finds size of objects.
    See https://goshippo.com/blog/measure-real-size-any-python-object/
    """
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_size(v, seen) for v in obj.values()])
        size += sum([get_size(k, seen) for k in obj.keys()])
    elif hasattr(obj, '__dict__'):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen) for i in obj])
    return size

def minimum_string_match(s,valid_strings):
    """return the valid string given a minimum string input"""
    pass

def uniq(seq):
    """ Remove duplicates from a list while preserving order.
        from http://stackoverflow.com/questions/480214/how-do-you-remove-duplicates-from-a-list-in-python-whilst-preserving-order 
    """
    seen = set()
    seen_add = seen.add
    return [ x for x in seq if x not in seen and not seen_add(x)]




"""Load generic SDFITS files
    - Not typically used directly.  Sub-class for specific telescope SDFITS flavors.
"""
import sys
import copy
from astropy.wcs import WCS
from astropy.units import cds
from astropy.io import fits
from astropy.modeling import models, fitting
import astropy.units as u
from astropy.table import Table
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from ..spectra.spectrum import Spectrum
from ..spectra.obsblock import Obsblock
from ..spectra import dcmeantsys
from ..util import uniq


class SDFITSLoad(object):
    '''
    Generic Container for a bintable(s) from selected HDU(s)

    Parameters
    -----------
    filename: - str 
        input file name
    source: - str
        target source to select from input file. Default: all sources
    hdu: - int or list
        Header Data Unit to select from input file. Default: all HDUs
    '''
    def __init__(self, filename, source=None, hdu=None, **kwargs):
        print("==SDFITSLoad %s" % filename)
        cds.enable()  # to get mmHg
        kwargs_opts = {'fix':False}
        kwargs_opts = {'wcs':False}
        kwargs_opts.update(kwargs)
        self._filename = filename
        self._bintable = []
        self._ptable = []
        self._binheader = []
        self._data = []
        self._obsblock = [] # list of SpecList
        self._hdu = fits.open(filename)  
        self._header = self._hdu[0].header
        self.load(hdu,**kwargs_opts)
        #self.create_index()
        #self._hdu.close()  # can't access hdu[i].data member of you do this.

    @property 
    def filename(self):
        return self._filename
    
    def reset(self,hdu=None):
        self._bintable = []
        self._binheader = []
        self._ptable = []
        self._data = []
        self._spectra = []
        self._hdu = fits.open(self._filename)  
        self._header = self._hdu[0].header
        #if hdu is not None:
         #   self.load(src,hdu)
    
    def create_index(self,hdu=None):
        if hdu is not None:
            ldu = list([hdu])
        else:
            ldu = range(1,len(self._hdu))
        self._ptable = []
        for i in ldu:
            t = Table.read(self._hdu[i]) 
            t.remove_column('DATA')
            self.stripT(t)
            print(f"doing pandas for HDU {i}")
            self._ptable.append(t.to_pandas())
            del t
            
    def load(self, hdu=None, **kwargs):
        """
        for given hdu make this bintable available
        Note mmHg and UTC are unrecognized units.  mmHg is in astropy.units.cds but UTC is just wrong.
        """

        self._bintable = []
        self._ptable = []
        self._binheader = []
        self._data = []
        self._nrows = []
        source = kwargs.get('source',None)
        fix = kwargs.get('fix',False)
        wcs = kwargs.get('wcs',False)

        if hdu is not None:
            ldu = list([hdu])
        else:
            ldu = range(1,len(self._hdu))
        for i in ldu:
            j=i-1
            self._bintable.append(self._hdu[i]) 
            self._binheader.append(self._hdu[i].header)
            # TODO: don't allow preselection here, it just screws bookkeepingu p.
            # All selection must happen in obsblocks.
            #if source is None:
            #self._data.append(self._bintable[j].data[:]["DATA"])
            #else:
            #    wh1 = np.char.strip(self._bintable[j]['OBJECT']) == source # true/false array
            #    if wh1.sum() == 0: #all False none found
            #        srcs = np.unique(self._bintable[j]['OBJECT'])
            #        raise Exception(f"Source name {source} not found in HDU {i}. Sources present are {srcs}")
            #    self._data.append(self._bintable[j]["DATA"][wh1])
            #    mask = np.where(wh1)[0]
            #    self._bintable[j] = self._bintable[j][mask]  # header will be wrong?
            self._nrows.append(self._binheader[j]["NAXIS2"])

        if kwargs.get("index",False):
            self.create_index(hdu)

    def stripT(self,b):
        # remove leading and trailing chars from all strings in table 
        for n in b.colnames:
            if np.issubdtype(b.dtype[n],str):
                b[n] = np.char.strip(b[n])

    def strip(self):
        # remove leading and trailing chars from all strings in table 
        for b in self._ptable:
            for n in b.colnames:
                if np.issubdtype(b.dtype[n],str):
                    b[n] = np.char.strip(b[n])

    def _loadlists(self,hdu,fix=False,wcs=False,maxspect=1E16):
        self._obsblock = []
        i=0
        k = -1
        print("HDU = ",hdu)
        if hdu is not None:
            b = self._ptable[hdu-1]
            rawspect = self._bintable[i].data["DATA"]
            sl = SpectrumList()
            maxload = int(np.min([maxspect,self.nrows(i)]))
            print(f"Creating {maxload} Spectrum in bintable {i} HDU {hdu}",file=sys.stderr)
            for j in range(maxload):
                k = k+1
                # need extra [[]] because we have 1x1 spatial NAXIS
                # otherwise, slicing the spectrum won't work.
                if wcs: 
                    sp = np.array([[self.rawspectrum(i,j)]])
                else:
                    #sp = self.rawspectrum(i,j)*u.K
                    sp = np.copy(rawspect[j])#*u.K
                    #sp = np.random.rand(32768)*u.K
                naxis1 =  sp.shape[0]#self.nchan(i)
                printme = int(0.1*len(b))
                if (k%printme) == 0: 
                    print(f"Row {k} nchan {naxis1} {type(sp)}", file=sys.stderr)
                    #print(f"NAXIS1 is {naxis1}",file=sys.stderr)
                crval1  = b['CRVAL1'][j]
                cdelt1  = b['CDELT1'][j]
                crpix1  = b['CRPIX1'][j]
                ctype1  = b['CTYPE1'][j]
                # Ensure rest frequency is in Hertz
                # CUNIT1 is not always present
                restfrq = b['RESTFREQ'][j]
                if "CUNIT1" in b.columns:
                    cunit1  = b['CUNIT1'][j]
                    rfq = restfrq * u.Unit(cunit1)
                    restfreq = rfq.to("Hz").value
                cunit1 = "Hz"
                crval2  = b['CRVAL2'][j]
                crval3  = b['CRVAL3'][j]
                ctype2  = b['CTYPE2'][j]
                ctype3  = b['CTYPE3'][j]
                # 'FREQ-OBS' to 'FREQ'; assuming SPECSYS='TOPOCENT'
                #if ctype1 == 'FREQ-OBS': ctype1  = 'FREQ'
                # only axis1 needs a full description, axis2,3,4 are all single points
                if wcs:
                    wcs = WCS(header={'CDELT1': cdelt1, 'CRVAL1': crval1, 'CUNIT1': cunit1,
                                      'CTYPE1': 'FREQ', 'CRPIX1': crpix1, 'RESTFRQ': restfrq,
                                      'CTYPE2': ctype2, 'CRVAL2': crval2, 'CRPIX2': 1,
                                      'CTYPE3': ctype3, 'CRVAL3': crval3, 'CRPIX3': 1,
                                      'CUNIT2': 'deg', 'CUNIT3':'deg',
                                      'NAXIS1': naxis1, 'NAXIS2':1, 'NAXIS3':1
                                     },
                             fix=fix)
                else:
                    wcs = None
                # GBT really fucks up FREQ/VELDEF/VELFRAME
                #if False:
                if "VELFRAME" in b.columns:
                    vframe = b['VELFRAME'][j]
                elif "VFRAME" in b.columns:
                    vframe = b['VFRAME'][j]
                else:
                    vframe = None
                if "VELDEF" in b.columns:
                    vdef = b['VELDEF'][j]
                else:
                    vdef = None

                meta={'CDELT1': cdelt1, 'CRVAL1': crval1, 'CUNIT1': cunit1,
                      'CTYPE1': 'FREQ', 'CRPIX1': crpix1, 'RESTFRQ': restfrq,
                      'CTYPE2': ctype2, 'CRVAL2': crval2, 'CRPIX2': 1,
                      'CTYPE3': ctype3, 'CRVAL3': crval3, 'CRPIX3': 1,
                      'CUNIT2': 'deg', 'CUNIT3':'deg',
                      'NAXIS1': naxis1, 'NAXIS2':1, 'NAXIS3':1,
                      'VELDEF': vdef, 'VELFRAME': vframe
                     }

                convention = self.velocity_convention(vdef,vframe)
                #meta = dict(b.loc[j])# Necessary? Since we are sending whole pandas table to Obsblock
                if fix:
                    self.fix_meta(meta)
                if False:
                    try:
                        convention = self.velocity_convention(meta['VELDEF'],meta['VELFRAME'])
                    except Exception:
                        #print("WARNING: insufficient veldef/velframe, assuming convention is 'doppler_radio'")
                        convention="doppler_radio"
                meta = {}
                sl.append(Spectrum(flux=sp*u.K,wcs=wcs,  meta=meta, velocity_convention=convention))
            self._obsblock.append(Obsblock(sl,self._ptable[i]))
            i=i+1


    def fix_meta(self,meta):
        """Do any repair to the meta/header for peculariaties in definitions from a particular observatory
        The passed-in dictionary will be repaired in place.
        At minimum this method must populate meta['VELDEF'] and meta['VELFRAME']
        """
        pass
    
    def velocity_convention(self,veldef,velframe):
        # sub-classes must implement this so I can keep this class generic.
        # GBT uses VELDEF and VELFRAME incorrectly. 
        return "doppler_radio"
    
    def udata(self,bintable,key):
        return utils.uniq(self._data[bintable][key])
                                  
    def ushow(self,bintable,key):
        print(f'{bintable} {key}: {self.udata(bintable,key)}')
        
    def naxis(self,bintable,naxis):
        nax = f'NAXIS{naxis}'
        return self._binheader[bintable][nax]
    
    def nintegrations(self,bintable,source=None):
        if source is not None:
            nint = np.shape(np.char.strip(self._data[bintable]['OBJECT']) == source)[0]//self.npol(bintable)
        else:
            nint = np.shape(self._data[bintable])[0]//self.npol(bintable)
        return nint

    def rawspectra(self,bintable):
        return self._bintable[bintable].data[:]["DATA"]

    def rawspectrum(self,bintable,i):
        return self._bintable[bintable].data[:]["DATA"][i]
    
    def nrows(self,bintable):
        return self._nrows[bintable]

    def nchan(self,bintable):
        return np.shape(self.rawspectrum(bintable,1))[0]
    
    def npol(self,bintable):
        return len(self.udata(bintable,'CRVAL4'))
    
    def sources(self,bintable):
        return self.udata(bintable,'OBJECT')
    
    def scans(self,bintable):
        return self.ushow(bintable,'SCAN')
    
    def __len__(self):
        return self.nrows
    
    def _summary(self,bintable):
        j=bintable
        nrows = self.naxis(j,2)
        nflds = self._binheader[j]['TFIELDS']
        restfreq = np.unique(self._data[j]['RESTFREQ'])/1.0E9
    #
        print("HDU       %d" %  (j+1))
        print("BINTABLE: %d rows x %d cols with %d chans" % (self._nrows[j],nflds,self.nchan(j)))
        print("Selected  %d/%d rows" % (self._nrows[j],nrows))
        print("Sources: ",self.sources(j))
        print("RESTFREQ:",restfreq,'GHz')
        print("Scans:   ",self.scans(j))
        print("Npol:    ",self.npol(j))   
        print("Nint:    ",self.nintegrations(j))
        
    def summary(self):
        """Print a summary of each record of the data"""
        print("File:     %s"%self._filename)
        for i in range(len(self._bintable)):
            self._summary(i)
    
    def __repr__(self):
        return self._filename    

def sonoff(scan, procseqn):
    """
    return the list of On and Off scan numbers
    there must be a more elegant python way to do this....
    """
    sp = {}
    for (i,j) in zip(scan, procseqn):
        sp[i] = j
    
    us1 = utils.uniq(scan)
    up1 = utils.uniq(procseqn)
    
    sd = {}
    for i in up1:
        sd[i] = []
        
    for s in us1:
        sd[sp[s]].append(s)

    return sd

